﻿using System;
using System.Linq;
using System.Collections.ObjectModel;
using System.Windows.Forms;
using System.IO;

/*    Autoencoder jusing a neural net with backpropagation       */
/*                        www.mosismath.com                      */

namespace WindowsFormsApplication1
{
    

    public partial class Mainwin : Form
    {
        int samples = 0;
        const int features = 4;    // number of features (or dimensions) of the input data
        const int redfeatures = 2; // number of features (or dimensions) of the reduced data
      
        TData data = new TData();
        TAutoEncoder autoEncoder;
             
        public Mainwin()
        {
            int i, j;
            DataGridViewCell cell;

            InitializeComponent();
            data.readData(Application.StartupPath + "\\data.csv");

            autoEncoder = new TAutoEncoder(features, redfeatures, data);
            samples = data.values.Count;

            dgData.RowCount = samples;
            dgData.ColumnCount = features+1;
            for (j = 0; j < features; j++)
                dgData.Columns[j].HeaderText = "x" + (j+1).ToString();
            dgData.Columns[j].HeaderText = "y";
            for (i = 0; i < samples; i++)
            {
                dgData.Rows[i].HeaderCell.Value = (i+1).ToString();
    
                for (j = 0; j <= features; j++)
                {
                    cell = dgData[j, i];
                    cell.Value = Math.Round(data.values.ElementAt(i)[j], 4).ToString();
                }
            }
        }

        private void bt_Calc_Click(object sender, EventArgs e)
        {
            int i, j, k;
            Collection<double[]> reducedValues = new Collection<double[]>();

            data.calcMean(features);
            data.normalize(features);
 
            autoEncoder.Calculate();

            // build the reduced data    
            double[] tempVals;
            double[] res;
            for (k = 0; k < samples; k++)
            {
                double[] reducedLine = new double[redfeatures + 1];
                tempVals = data.values.ElementAt(k);
                res = autoEncoder.ForwardProp(tempVals);
                for (i = 0; i < redfeatures; i++)
                {
                    reducedLine[i] = 0;
                    for (j = 0; j < features; j++)
                        reducedLine[i] = res[i];
                }
                reducedLine[redfeatures] = tempVals[features];
                reducedValues.Add(reducedLine);
            }

            tbCost.Text = Math.Round(autoEncoder.cost, 8).ToString();
         
            for (k = 0; k < samples; k++)
            {
                string str = "";
                for (int h = 0; h < reducedValues.ElementAt(k).Length; h++)
                {
                    if (h < reducedValues.ElementAt(k).Length - 1)
                        str = str + reducedValues.ElementAt(k)[h].ToString() + ",";
                    else
                        str = str + reducedValues.ElementAt(k)[h].ToString();
                }
               
                File.AppendAllText("reduceddata.csv", str + "\n");
            }
            autoEncoder.writeTrainedData(Application.StartupPath + "\\autoencoderdata.json");
        }
    }
}
