﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;
using System.Text;

namespace WindowsFormsApplication1
{
    [DataContract]
    public struct TWLine
    {
        [DataMember]
        public double[] w;
    }

    [Serializable]
    public class TAutoEncoderData
    {
        [DataMember]
        public TWLine[] AutoEncoderLine;

        [DataMember]
        public double[] mean;

        [DataMember]
        public double[] deviation;
    }

    public class TAutoEncoder
    {
        int layers = 1;
        public int features;
        public int reduced;
        const int iterations = 2000000;
        double learning_rate = 0.00005;
        public double cost = 1000;
        double costSum = 0;
        int samples = 29;
        public List<TLayer> net = new List<TLayer>();
        TData data;

        public TAutoEncoder(int iFeatures, int reducedFeatures, TData sData)
        {
            TLayer tempRow;
            data = sData;
            features = iFeatures;
            reduced = reducedFeatures;
            tempRow = new TLayer(features, reducedFeatures);
            net.Add(tempRow);
            tempRow = new TLayer(reducedFeatures, features);
            net.Add(tempRow);
            layers = net.Count;
        }

        public double[] ForwardProp(double[] x)
        {
            int i, j, k;
            TLayer actLayer = net.ElementAt(0);
            double[] actX = new double[x.Length];
            for (i = 0; i < x.Length; i++)
                actX[i] = x[i];

            for (i = 0; i < layers; i++)
            {
                for (j = 0; j < actLayer.featuresOut; j++)
                {
                    actLayer.x[j] = 0;
                    for (k = 0; k < actLayer.featuresIn; k++)
                    {
                        actLayer.x[j] = actLayer.x[j] + actX[k] * actLayer.w[k, j];
                    }
                    actLayer.o[j] = actLayer.Act(actLayer.x[j]);
                }

                for (j = 0; j < actLayer.x.Length; j++)
                    actX[j] = actLayer.o[j];

                if (i < layers - 1)
                    actLayer = net.ElementAt(i + 1);
            }
            return net.ElementAt(0).x;
        }


        private void BackwardProp(double[] x, double[] y)
        {
            int i, j, k;
            TLayer rightLayer = net.ElementAt(layers - 1);   // actual computed layer
            TLayer leftLayer = net.ElementAt(0);   // actual computed layer
            double[] actX = new double[x.Length];

            for (j = 0; j < leftLayer.x.Length; j++)
                actX[j] = leftLayer.x[j];
            for (j = 0; j < rightLayer.featuresOut; j++)
            {
                costSum = costSum + ((y[j] - rightLayer.o[j]) * (y[j] - rightLayer.o[j]));
            }
           
            for (i = 0; i < rightLayer.featuresIn; i++)
            {
                for (k = 0; k < rightLayer.featuresOut; k++)
                {
                    rightLayer.deltaW[i, k] += (rightLayer.w[i, k] * x[k] + leftLayer.o[i]) * (y[k] - rightLayer.o[k]);
                }
            }
        }
  

        private void UpdateValues()
        {
            int k, l;
            TLayer rightLayer = net.ElementAt(layers - 1);
            TLayer lefttLayer = net.ElementAt(0);

            for (k = 0; k < rightLayer.featuresOut; k++)
            {
                for (l = 0; l < rightLayer.featuresIn; l++)
                {
                    rightLayer.w[l, k] = rightLayer.w[l, k] + rightLayer.deltaW[l, k] / samples * learning_rate;
                    // copy the values of the right layer to the left layer
                    lefttLayer.w[k, l] = rightLayer.w[l, k];
                }
            }
        }


        public void Calculate()
        {
            int i, j, k;
            int count = 0;
            double deltaCost = 10000;
            double[] tempX = new double[features];
            double[] tempY = new double[features];
            TLayer actLayer = net.ElementAt(0);

            while ((count < iterations) && (deltaCost > 1e-15))
            {
                actLayer = net.ElementAt(0);

                for (i = 0; i < layers; i++)
                {
                    actLayer = net.ElementAt(i);
                    for (j = 0; j < actLayer.featuresOut; j++)
                    {
                        for (k = 0; k < actLayer.featuresIn; k++)
                        {
                            actLayer.deltaW[k, j] = 0;
                        }
                    }
                }
                count++;

                costSum = 0;
                for (i = 0; i < samples; i++)
                {
                    for (k = 0; k < features; k++)
                    {
                        tempX[k] = data.values.ElementAt(i)[k];
                        tempY[k] = tempX[k];
                    }
                    ForwardProp(tempX);
                    BackwardProp(tempX, tempY);
                }
                UpdateValues();
                deltaCost = Math.Abs(cost - costSum / samples);
                cost = costSum / samples;
            }
        }

        public double[,] GetParameters()
        {
            return net.ElementAt(1).w;
        }
    
        public void writeTrainedData(string FileName)
        {
            int i, j;
            TAutoEncoderData parameter = new TAutoEncoderData();
            parameter.AutoEncoderLine = new TWLine[features];
            MemoryStream ms = new MemoryStream();
            DataContractJsonSerializer ser = new DataContractJsonSerializer(typeof(TAutoEncoderData));

            parameter.mean = data.mean;
            parameter.deviation = data.deviation;
            for (i = 0; i < features; i++)
            {
                parameter.AutoEncoderLine[i].w = new double[reduced];
                for (j = 0; j < reduced; j++)
                {
                    parameter.AutoEncoderLine[i].w[j] = net.ElementAt(0).w[i, j];
                }        
            }
            ser.WriteObject(ms, parameter);
        
            using (FileStream fs = new FileStream(FileName, FileMode.Create))
            {
                ms.Position = 0;
                fs.Write(ms.ToArray(), 0, ms.ToArray().Length);
                ms.Close();
                fs.Close();
            }
        }


        public void readTrainedData(string FileName)
        {
            int i;
            using (StreamReader r = new StreamReader(FileName))
            {
                string response = r.ReadToEnd();
                TAutoEncoderData parameter = new TAutoEncoderData();
                parameter.AutoEncoderLine = new TWLine[features];
                for (i = 0; i < features; i++)
                {
                    parameter.AutoEncoderLine[i].w = new double[reduced];
                }
                parameter.mean = new double[features];
                MemoryStream m = new MemoryStream(Encoding.UTF8.GetBytes(response));
                DataContractJsonSerializer ser = new DataContractJsonSerializer(parameter.GetType());
                parameter = ser.ReadObject(m) as TAutoEncoderData;
                data.mean = parameter.mean;
                for (i = 0; i < features; i++)
                {
                    parameter.AutoEncoderLine[i].w = new double[reduced];
                }
            }
        }
    }
}
